from dataclasses import dataclass
from typing import Dict, List, Optional

import yaml
from pydantic import BaseModel


@dataclass
class SageMakerMetricsTrackingConfig:
    metrics: dict


class ResourceConfig(BaseModel):
    instance_type: str
    instance_count: Optional[int]
    timeout_seconds: Optional[int]
    security_group_ids: Optional[List[str]]
    subnets: Optional[List[str]]


class SageMakerConfig(BaseModel):
    pipeline_names_mapping: Optional[Dict[str, str]] = None


class DockerConfig(BaseModel):
    image: str
    working_directory: str = "/home/kedro"


class AwsConfig(BaseModel):
    execution_role: str
    bucket: str
    sagemaker: SageMakerConfig
    resources: Dict[str, ResourceConfig]


class KedroSageMakerPluginConfig(BaseModel):
    aws: AwsConfig
    docker: DockerConfig


CONFIG_TEMPLATE_YAML = """
aws:
  # Bucket name to use as a temporary storage within the pipeline job
  bucket: "{bucket}"

  # AWS SageMaker Executor role ARN
  execution_role: "{execution_role}"

  # use Kedro node tags (recommended) or node names to assign compute resources
  # use __default__ to specify the default values (for all nodes)
  resources:
    __default__:
      instance_count: 1
      instance_type: ml.m5.large
      timeout_seconds: 86400
      security_group_ids: null
      subnets: null
  sagemaker:
    # (optional) mapping between kedro pipeline names (keys) and SageMaker pipeline names
    # Note that SageMaker does not support underscores in pipeline names.
    # Here you can map for example add `__default__: "my-pipeline"`
    # to make the `__default__` Kedro pipeline appear as `my-pipeline` in SageMaker UI
    pipeline_names_mapping:
      kedro_pipeline_name: "sagemaker-pipeline-name"
docker:
  image: "{docker_image}"
  working_directory: /home/kedro
""".strip()

# This auto-validates the template above during import
_CONFIG_TEMPLATE = KedroSageMakerPluginConfig.parse_obj(
    yaml.safe_load(CONFIG_TEMPLATE_YAML)
)



import json
from itertools import chain
from types import SimpleNamespace
from typing import Dict, Iterator, List, Optional, Tuple, Union

from kedro.framework.context import KedroContext
from kedro.io import DataCatalog
from kedro.pipeline import Pipeline as KedroPipeline
from kedro.pipeline.node import Node as KedroNode
from sagemaker import Model, Processor
from sagemaker.estimator import Estimator
from sagemaker.network import NetworkConfig
from sagemaker.workflow.execution_variables import ExecutionVariables
from sagemaker.workflow.model_step import ModelStep
from sagemaker.workflow.parameters import (
    ParameterBoolean,
    ParameterFloat,
    ParameterInteger,
    ParameterString,
)
from sagemaker.workflow.pipeline import Pipeline as SageMakerPipeline
from sagemaker.workflow.pipeline_context import (
    LocalPipelineSession,
    PipelineSession,
)
from sagemaker.workflow.steps import ProcessingStep, StepTypeEnum, TrainingStep

from kedro_sagemaker.config import (
    KedroSageMakerPluginConfig,
    ResourceConfig,
    SageMakerMetricsTrackingConfig,
)
from kedro_sagemaker.constants import (
    KEDRO_SAGEMAKER_ARGS,
    KEDRO_SAGEMAKER_EXECUTION_ARN,
    KEDRO_SAGEMAKER_METRICS,
    KEDRO_SAGEMAKER_PARAM_KEY_PREFIX,
    KEDRO_SAGEMAKER_PARAM_VALUE_PREFIX,
    KEDRO_SAGEMAKER_RUNNER_CONFIG,
    KEDRO_SAGEMAKER_WORKING_DIRECTORY,
)
from kedro_sagemaker.datasets import SageMakerModelDataset
from kedro_sagemaker.runner import KedroSageMakerRunnerConfig
from kedro_sagemaker.utils import flatten_dict, is_mlflow_enabled

SageMakerStepType = Union[ProcessingStep, TrainingStep, ModelStep]


class KedroSageMakerGenerator:
    def __init__(
        self,
        pipeline_name: str,
        kedro_context: KedroContext,
        config: KedroSageMakerPluginConfig,
        docker_image: Optional[str] = None,
        is_local: bool = False,
        execution_role_arn: Optional[str] = None,
    ):
        self.is_local = is_local
        self.docker_image = docker_image or config.docker.image
        self.config = config
        self.kedro_context = kedro_context
        self.pipeline_name = pipeline_name
        self.execution_role_arn = execution_role_arn

    @property
    def _execution_role(self):
        return self.execution_role_arn or self.config.aws.execution_role

    def _datasets_of_type(
        self, t: type, inputs_or_outputs: List[str], catalog: DataCatalog
    ) -> Iterator[str]:
        defined_datasets = set(catalog.list())
        for ds in inputs_or_outputs:
            if ds in defined_datasets and isinstance(catalog._get_dataset(ds), t):
                yield ds

    def _prepare_sagemaker_params(self) -> Tuple[list, dict]:
        sm_param_types = {
            int: ParameterInteger,
            float: ParameterFloat,
            bool: ParameterBoolean,
            str: ParameterString,
        }

        sm_param_envs = {}
        sm_kedro_params = []
        for node_idx, (k, v) in enumerate(
            flatten_dict(self.kedro_context.params).items()
        ):
            sm_param_key = ParameterString(f"param{node_idx}", default_value=k)
            value_name = f"value{node_idx}"

            if (t := type(v)) in sm_param_types:
                sm_param_value = sm_param_types[t](value_name, default_value=v)
            else:
                sm_param_value = ParameterString(
                    value_name, default_value=json.dumps(v)
                )

            sm_kedro_params.append(sm_param_key)
            sm_kedro_params.append(sm_param_value)

            sm_param_envs[
                KEDRO_SAGEMAKER_PARAM_KEY_PREFIX + str(node_idx)
            ] = sm_param_key
            sm_param_envs[KEDRO_SAGEMAKER_PARAM_VALUE_PREFIX + str(node_idx)] = (
                sm_param_value
                if isinstance(sm_param_value, ParameterString)
                else sm_param_value.to_string()
            )
        return sm_kedro_params, sm_param_envs

    def get_kedro_pipeline(self) -> KedroPipeline:
        from kedro.framework.project import pipelines

        pipeline: KedroPipeline = pipelines[self.pipeline_name]
        return pipeline

    def _get_sagemaker_pipeline_name(self) -> str:
        return (self.config.aws.sagemaker.pipeline_names_mapping or {}).get(
            self.pipeline_name, "kedro-sagemaker-pipeline"
        )

    def _get_default_resources(self) -> ResourceConfig:
        return ResourceConfig(instance_type="ml.m5.large")

    def _get_resources_for_node(self, node: Union[KedroNode, SimpleNamespace]):
        node_resources = next(
            (
                self.config.aws.resources.get(n)
                for n in chain([node.name], iter(node.tags))
            ),
            None,
        )

        defined_default = self.config.aws.resources.get("__default__", None)
        defaults = self._get_default_resources()
        if defined_default:
            defaults = defaults.copy(update=defined_default.dict())

        if node_resources:
            return defaults.copy(
                update={key: val for key, val in node_resources if val is not None}
            )
        else:
            return defaults

    def generate(self) -> SageMakerPipeline:
        runner_config = KedroSageMakerRunnerConfig(bucket=self.config.aws.bucket)

        sagemaker_session = (
            LocalPipelineSession() if self.is_local else PipelineSession()
        )

        sm_params = []

        sm_params_kedro, sm_param_envs = self._prepare_sagemaker_params()
        sm_params.extend(sm_params_kedro)

        steps: Dict[str, SageMakerStepType] = {}
        sm_training_steps_with_model_outputs = {}

        pipeline = self.get_kedro_pipeline()

        for node_idx, node in enumerate(pipeline.nodes):
            node_resources = self._get_resources_for_node(node)
            sm_node_name = node.name.replace(".", "__")

            sagemaker_model_inputs, sagemaker_model_outputs = self._get_model_io(node)

            has_sagemaker_model_output = len(sagemaker_model_outputs) > 0

            is_training_node = (
                hasattr(node.func, KEDRO_SAGEMAKER_METRICS)
                or has_sagemaker_model_output
            )

            if is_training_node:
                sm_metrics = getattr(
                    node.func,
                    KEDRO_SAGEMAKER_METRICS,
                    SageMakerMetricsTrackingConfig({}),
                )

                step = self._create_training_step(
                    node,
                    node_resources,
                    runner_config,
                    sagemaker_model_inputs,
                    sm_metrics,
                    sm_node_name,
                    sm_param_envs,
                    sm_training_steps_with_model_outputs,
                )

                assert (
                    len(sagemaker_model_outputs) <= 1
                ), "There can only be 1 SageMaker Model output per Kedro node!"
                if has_sagemaker_model_output:
                    model_steps = self._create_model_register_steps(
                        node_idx,
                        sagemaker_model_outputs,
                        sagemaker_session,
                        sm_training_steps_with_model_outputs,
                        step,
                    )
                    for ms in model_steps:
                        steps[f"{node.name}-{ms.name}"] = ms

            else:
                step = self._create_processing_step(
                    node,
                    node_resources,
                    runner_config,
                    sm_node_name,
                    sm_param_envs,
                )
            steps[node.name] = step

        self._add_step_dependencies(pipeline, steps)
        if is_mlflow_enabled():
            self._add_mlflow_support(steps, runner_config)

        smp = SageMakerPipeline(
            self._get_sagemaker_pipeline_name(),
            sagemaker_session=sagemaker_session,
            steps=list(steps.values()),
            parameters=sm_params,
        )
        return smp

    def _add_mlflow_support(self, steps, runner_config):
        mlflow_node_namespace = SimpleNamespace(name="start-mlflow-run", tags=[])
        mlflow_start_run = self._create_processing_step(
            node=mlflow_node_namespace,
            node_resources=self._get_resources_for_node(mlflow_node_namespace),
            runner_config=runner_config,
            sm_node_name="start-mlflow-run",
            sm_param_envs={},
            entrypoint=[
                "kedro",
                "sagemaker",
                "-e",
                self.kedro_context.env or "local",
                "mlflow-start",
            ],
        )
        for step in steps.values():
            if step.depends_on is not None:
                continue
            if step.step_type not in (StepTypeEnum.TRAINING, StepTypeEnum.PROCESSING):
                continue

            step.add_depends_on([mlflow_start_run])

        steps["start-mlflow-run"] = mlflow_start_run

    def _add_step_dependencies(
        self, pipeline, steps: Dict[str, SageMakerStepType]
    ) -> Dict[str, SageMakerStepType]:
        for node, dependencies in pipeline.node_dependencies.items():
            deps_names = set(d.name for d in dependencies)
            sm_step = steps[node.name]
            sm_deps = [s for n, s in steps.items() if n in deps_names]
            sm_step.add_depends_on(sm_deps)
        return steps

    def _create_processing_step(
        self,
        node,
        node_resources,
        runner_config,
        sm_node_name,
        sm_param_envs,
        entrypoint=None,
    ):
        step = ProcessingStep(
            sm_node_name,
            processor=Processor(
                entrypoint=entrypoint or self._get_kedro_command(node),
                role=self._execution_role,
                image_uri=self.docker_image,
                instance_count=node_resources.instance_count,
                instance_type=node_resources.instance_type,
                max_runtime_in_seconds=node_resources.timeout_seconds,
                env={
                    KEDRO_SAGEMAKER_RUNNER_CONFIG: runner_config.json(),
                    KEDRO_SAGEMAKER_EXECUTION_ARN: ExecutionVariables.PIPELINE_EXECUTION_ARN,
                    **sm_param_envs,
                },
                network_config=NetworkConfig(
                    security_group_ids=node_resources.security_group_ids,
                    subnets=node_resources.subnets,
                ),
            ),
            display_name=node.name,
            description=node.name,
        )
        return step

    def _get_kedro_command(self, node, as_string=False) -> Union[str, List[str]]:
        cmd = [
            "kedro",
            "sagemaker",
            "-e",
            self.kedro_context.env or "local",
            "execute",
            f"--pipeline={self.pipeline_name}",
            f"--node={node.name}",
        ]
        if as_string:
            return " ".join(cmd)
        else:
            return cmd

    def _create_model_register_steps(
        self,
        node_idx,
        sagemaker_model_outputs,
        sagemaker_session,
        sm_training_steps_with_model_outputs,
        step,
    ) -> Union[Tuple[ModelStep], Tuple[ModelStep, ModelStep]]:
        model_output_name = sagemaker_model_outputs[0].replace(".", "__")
        model = Model(
            image_uri=self.docker_image,
            model_data=step.properties.ModelArtifacts.S3ModelArtifacts,
            role=self._execution_role,
            name=model_output_name,
            sagemaker_session=sagemaker_session,
        )
        step_model_create = ModelStep(
            f"CreateModel{node_idx}",
            step_args=model.create(),
        )
        sm_training_steps_with_model_outputs[sagemaker_model_outputs[0]] = step
        results = (step_model_create,)

        if not self.is_local:
            model_registry_kwargs = dict(
                model_package_group_name="KedroSageMakerModels",
                content_types=["application/json"],
                response_types=["application/json"],
                domain="MACHINE_LEARNING",
                task="OTHER",
                image_uri=self.docker_image,
            )

            # TODO - maybe model metrics from https://docs.aws.amazon.com/sagemaker/latest/dg/define-pipeline.html ?
            ds: SageMakerModelDataset = self.kedro_context.catalog._get_dataset(
                sagemaker_model_outputs[0]
            )  # noqa

            model_registry_kwargs.update(ds.model_registry_kwargs)

            step_model_register = ModelStep(
                f"RegisterModel{node_idx}",
                step_args=model.register(**model_registry_kwargs),
            )

            results = results + (step_model_register,)
        return results

    def _create_training_step(
        self,
        node,
        node_resources,
        runner_config,
        sagemaker_model_inputs,
        sm_metrics,
        sm_node_name,
        sm_param_envs,
        sm_training_steps_with_model_outputs,
    ):
        return TrainingStep(
            sm_node_name,
            estimator=Estimator(
                image_uri=self.docker_image,
                role=self._execution_role,
                instance_count=node_resources.instance_count,
                instance_type=node_resources.instance_type,
                max_run=node_resources.timeout_seconds,
                metric_definitions=[
                    {"Name": k, "Regex": v} for k, v in sm_metrics.metrics.items()
                ],
                enable_sagemaker_metrics=len(sm_metrics.metrics) > 0,
                environment={
                    KEDRO_SAGEMAKER_ARGS: self._get_kedro_command(node, as_string=True),
                    KEDRO_SAGEMAKER_RUNNER_CONFIG: runner_config.json(),
                    KEDRO_SAGEMAKER_WORKING_DIRECTORY: self.config.docker.working_directory,
                    KEDRO_SAGEMAKER_EXECUTION_ARN: ExecutionVariables.PIPELINE_EXECUTION_ARN,
                    # "PYTHONPATH": "/home/kedro/src",
                    # # TODO - this will not be needed if plugin is installed, I hope :D,
                    **sm_param_envs,
                },
                base_job_name=sm_node_name,
                model_uri=sm_training_steps_with_model_outputs[
                    sagemaker_model_inputs[0]
                ].properties.ModelArtifacts.S3ModelArtifacts
                if sagemaker_model_inputs
                else None,
                security_group_ids=node_resources.security_group_ids,
                subnets=node_resources.subnets,
            ),
            display_name=node.name,
            description=node.name,
        )

    def _get_model_io(self, node):
        sagemaker_model_inputs = list(
            self._datasets_of_type(
                SageMakerModelDataset, node.inputs, self.kedro_context.catalog
            )
        )
        sagemaker_model_outputs = list(
            self._datasets_of_type(
                SageMakerModelDataset, node.outputs, self.kedro_context.catalog
            )
        )
        return sagemaker_model_inputs, sagemaker_model_outputs



# Taken from https://github.com/kedro-org/kedro-plugins/tree/main/kedro-docker
# and modified on 2022.12.13

DOCKERFILE_TEMPLATE = """
ARG BASE_IMAGE=python:3.9
FROM $BASE_IMAGE

# install project requirements
COPY src/requirements.txt /tmp/requirements.txt
RUN pip install -r /tmp/requirements.txt --no-cache-dir && rm -f /tmp/requirements.txt

WORKDIR /home/kedro
COPY . .

# Do not change the default entrypoint, it will break the Kedro SageMaker integration!
ENTRYPOINT ["kedro", "sagemaker", "entrypoint"]
""".lstrip()

DOCKERIGNORE_TEMPLATE = """
##########################
# Kedro PROJECT

# ignore Dockerfile and .dockerignore
Dockerfile
.dockerignore

# ignore potentially sensitive credentials files
conf/**/*credentials*

# ignore all local configuration
conf/local
!conf/local/.gitkeep

# ignore everything in the following folders
data
logs
notebooks
references
results

# except the following
!logs/.gitkeep
!notebooks/.gitkeep
!references/.gitkeep
!results/.gitkeep
!data/01_raw
""".lstrip()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['kedro_sagemaker']

package_data = \
{'': ['*']}

install_requires = \
['backoff>=2.2.1,<3.0.0',
 'cloudpickle>=2.2.0,<3.0.0',
 'kedro>=0.18.3,<0.19',
 'pydantic>=1.10.2,<2.0.0',
 's3fs>=2022.11.0,<2023.0.0',
 'sagemaker>=2.117.0,<3.0.0',
 'tarsafe==0.0.4',
 'zstandard>=0.19.0,<0.20.0']

entry_points = \
{'kedro.project_commands': ['sagemaker = kedro_sagemaker.cli:commands']}

setup_kwargs = {
    'name': 'kedro-sagemaker',
    'version': '0.3.0',
    'description': 'Kedro plugin with AWS SageMaker Pipelines support',
    'long_description': '# Kedro SageMaker Pipelines plugin\n\n[![Python Version](https://img.shields.io/pypi/pyversions/kedro-sagemaker)](https://github.com/getindata/kedro-sagemaker)\n[![License](https://img.shields.io/badge/license-Apache%202.0-blue.svg)](https://opensource.org/licenses/Apache-2.0)\n[![SemVer](https://img.shields.io/badge/semver-2.0.0-green)](https://semver.org/)\n[![PyPI version](https://badge.fury.io/py/kedro-sagemaker.svg)](https://pypi.org/project/kedro-sagemaker/)\n[![Downloads](https://pepy.tech/badge/kedro-sagemaker)](https://pepy.tech/project/kedro-sagemaker)\n\n[![Maintainability Rating](https://sonarcloud.io/api/project_badges/measure?project=getindata_kedro-sagemaker&metric=sqale_rating)](https://sonarcloud.io/summary/new_code?id=getindata_kedro-sagemaker)\n[![Coverage](https://sonarcloud.io/api/project_badges/measure?project=getindata_kedro-sagemaker&metric=coverage)](https://sonarcloud.io/summary/new_code?id=getindata_kedro-sagemaker)\n[![Documentation Status](https://readthedocs.org/projects/kedro-sagemaker/badge/?version=latest)](https://kedro-sagemaker.readthedocs.io/en/latest/?badge=latest)\n\n<p align="center">\n  <a href="https://getindata.com/solutions/ml-platform-machine-learning-reliable-explainable-feature-engineering"><img height="150" src="https://getindata.com/img/logo.svg"></a>\n  <h3 align="center">We help companies turn their data into assets</h3>\n</p>\n\n## About\nThis plugin enables you to run Kedro projects on Amazon SageMaker. Simply install the package and use the provided `kedro sagemaker` commands to build, push, and run your project on SageMaker.\n\n<img src="./docs/images/sagemaker_running_pipeline.gif" alt="Kedro SageMaker plugin" title="Kedro SageMaker plugin" />\n\n\n## Documentation \n\nFor detailed documentation refer to https://kedro-sagemaker.readthedocs.io/\n\n## Usage guide\n\n```\nUsage: kedro sagemaker [OPTIONS] COMMAND [ARGS]...\n\nOptions:\n  -e, --env TEXT  Environment to use.\n  -h, --help      Show this message and exit.\n\nCommands:\n  compile  Compiles the pipeline to a JSON file\n  init     Creates basic configuration for Kedro SageMaker plugin\n  run      Runs the pipeline on SageMaker Pipelines\n```\n\n## Quickstart\nFollow **quickstart** section on [kedro-sagemaker.readthedocs.io](https://kedro-sagemaker.readthedocs.io/) to see how to run your Kedro project on AWS SageMaker or watch the video below:\n\n<a href="https://www.youtube.com/watch?v=yXIdz4kNMc8">\n    <img src="./docs/images/kedro-sagemaker-video-tutorial.jpg" alt="Kedro SageMaker video tutorial" title="Kedro SageMaker video tutorial" />\n</a>\n',
    'author': 'Marcin Zabłocki',
    'author_email': 'marcin.zablocki@getindata.com',
    'maintainer': 'GetInData MLOPS',
    'maintainer_email': 'mlops@getindata.com',
    'url': 'https://github.com/getindata/kedro-sagemaker',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)
